from .api import Api
from .handlers import scheduled, stream, task, partial_rerun_merge
from .logger import CORVA_LOGGER as Logger
from .models.rerun import RerunDepth, RerunDepthRange, RerunTime, RerunTimeRange
from .models.scheduled.scheduled import (
    ScheduledDataTimeEvent,
    ScheduledDepthEvent,
    ScheduledNaturalTimeEvent,
)
from .models.stream.stream import (
    StreamDepthEvent,
    StreamDepthRecord,
    StreamTimeEvent,
    StreamTimeRecord,
)
from .models.merge.merge import PartialRerunMergeEvent
from .models.task import TaskEvent
from .service.cache_sdk import UserRedisSdk as Cache
from .shared import SECRETS as secrets


def __getattr__(name):
    import warnings

    if name == "ScheduledEvent":
        warnings.warn(
            "The corva.ScheduledEvent class is deprecated "
            "and will be removed from corva in the next major version. "
            "Import corva.ScheduledDataTimeEvent instead.",
            FutureWarning,
            stacklevel=2,
        )

        return ScheduledDataTimeEvent

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



import contextlib
import functools
import itertools
import logging
import sys
import warnings
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import pydantic
import redis

from corva.api import Api
from corva.configuration import SETTINGS
from corva.logger import CORVA_LOGGER, CorvaLoggerHandler, LoggingContext
from corva.models.base import RawBaseEvent
from corva.models.context import CorvaContext
from corva.models.merge.merge import PartialRerunMergeEvent
from corva.models.merge.raw import RawPartialRerunMergeEvent
from corva.models.scheduled.raw import RawScheduledEvent
from corva.models.scheduled.scheduled import ScheduledEvent, ScheduledNaturalTimeEvent
from corva.models.scheduled.scheduler_type import SchedulerType
from corva.models.stream.raw import RawStreamEvent
from corva.models.stream.stream import StreamEvent
from corva.models.task import RawTaskEvent, TaskEvent, TaskStatus
from corva.service import service
from corva.service.api_sdk import CachingApiSdk, CorvaApiSdk
from corva.service.cache_sdk import FakeInternalCacheSdk, InternalRedisSdk, UserRedisSdk

StreamEventT = TypeVar("StreamEventT", bound=StreamEvent)
ScheduledEventT = TypeVar("ScheduledEventT", bound=ScheduledEvent)
HANDLERS: Dict[Type[RawBaseEvent], Callable] = {}
GENERIC_APP_EVENT_TYPES = [RawStreamEvent, RawScheduledEvent, RawTaskEvent]


def get_cache_key(
    provider: str,
    asset_id: int,
    app_stream_id: int,
    app_key: str,
    app_connection_id: int,
) -> str:
    return (
        f"{provider}/well/{asset_id}/stream/{app_stream_id}/"
        f"{app_key}/{app_connection_id}"
    )


def base_handler(
    func: Callable,
    raw_event_type: Type[RawBaseEvent],
    handler: Optional[logging.Handler],
    merge_events: bool = False,
) -> Callable[[Any, Any], List[Any]]:
    @functools.wraps(func)
    def wrapper(aws_event: Any, aws_context: Any) -> List[Any]:
        with LoggingContext(
            aws_request_id=aws_context.aws_request_id,
            asset_id=None,
            app_connection_id=None,
            handler=logging.StreamHandler(stream=sys.stdout),
            user_handler=handler,
            logger=CORVA_LOGGER,
        ) as logging_ctx:
            (
                raw_custom_event_type,
                custom_handler,
            ) = _get_custom_event_type_by_raw_aws_event(aws_event)
            specific_callable = custom_handler or func

            try:
                context = CorvaContext.from_aws(
                    aws_event=aws_event, aws_context=aws_context
                )

                redis_client = redis.Redis.from_url(
                    url=SETTINGS.CACHE_URL, decode_responses=True, max_connections=1
                )
                data_transformation_type = raw_custom_event_type or raw_event_type
                if merge_events:
                    aws_event = _merge_events(aws_event, data_transformation_type)
                raw_events = data_transformation_type.from_raw_event(event=aws_event)

                if (
                    custom_handler is None
                    and data_transformation_type not in GENERIC_APP_EVENT_TYPES
                ):
                    CORVA_LOGGER.warning(
                        f"Handler for {data_transformation_type.__name__!r} "
                        f"event not found. Skipping..."
                    )
                    return []

                results = [
                    specific_callable(
                        raw_event,
                        context.api_key,
                        context.aws_request_id,
                        logging_ctx,
                        redis_client,
                    )
                    for raw_event in raw_events
                ]

                return results

            except Exception:
                CORVA_LOGGER.exception("The app failed to execute.")
                raise

    return wrapper


def stream(
    func: Optional[Callable[[StreamEventT, Api, UserRedisSdk], Any]] = None,
    *,
    handler: Optional[logging.Handler] = None,
    merge_events: bool = False,
) -> Callable:
    """Runs stream app.

    Arguments:
        handler: logging handler to include in Corva logger.
        merge_events: if True - merge all incoming events into one before
          passing them to func
    """

    if func is None:
        return functools.partial(stream, handler=handler, merge_events=merge_events)

    @functools.wraps(func)
    @functools.partial(
        base_handler,
        raw_event_type=RawStreamEvent,
        handler=handler,
        merge_events=merge_events,
    )
    def wrapper(
        event: RawStreamEvent,
        api_key: str,
        aws_request_id: str,
        logging_ctx: LoggingContext,
        redis_client: redis.Redis,
    ) -> Any:
        logging_ctx.asset_id = event.asset_id
        logging_ctx.app_connection_id = event.app_connection_id

        api = Api(
            api_url=SETTINGS.API_ROOT_URL,
            data_api_url=SETTINGS.DATA_API_ROOT_URL,
            api_key=api_key,
            app_key=SETTINGS.APP_KEY,
            timeout=None,
            app_connection_id=event.app_connection_id,
        )

        hash_name = get_cache_key(
            provider=SETTINGS.PROVIDER,
            asset_id=event.asset_id,
            app_stream_id=event.app_stream_id,
            app_key=SETTINGS.APP_KEY,
            app_connection_id=event.app_connection_id,
        )

        user_cache_sdk = UserRedisSdk(
            hash_name=hash_name, redis_dsn=SETTINGS.CACHE_URL, redis_client=redis_client
        )
        internal_cache_sdk = InternalRedisSdk(
            hash_name=hash_name, redis_client=redis_client
        )

        records = event.filter_records(
            old_max_record_value=event.get_cached_max_record_value(cache=user_cache_sdk)
        )

        if not records:
            # we've got the duplicate data if there are no records left after filtering
            return

        app_event = event.metadata.log_type.event.parse_obj(
            event.copy(update={"records": records}, deep=True)
        )
        with LoggingContext(
            aws_request_id=aws_request_id,
            asset_id=event.asset_id,
            app_connection_id=event.app_connection_id,
            handler=CorvaLoggerHandler(
                max_message_size=SETTINGS.LOG_THRESHOLD_MESSAGE_SIZE,
                max_message_count=SETTINGS.LOG_THRESHOLD_MESSAGE_COUNT,
                logger=CORVA_LOGGER,
                placeholder=" ...",
            ),
            user_handler=handler,
            logger=CORVA_LOGGER,
        ):
            result = service.run_app(
                has_secrets=event.has_secrets,
                app_key=SETTINGS.APP_KEY,
                api_sdk=CachingApiSdk(
                    api_sdk=CorvaApiSdk(api_adapter=api),
                    ttl=SETTINGS.SECRETS_CACHE_TTL,
                ),
                cache_sdk=internal_cache_sdk,
                app=functools.partial(
                    cast(Callable[[StreamEvent, Api, UserRedisSdk], Any], func),
                    app_event,
                    api,
                    user_cache_sdk,
                ),
            )

        try:
            event.set_cached_max_record_value(cache=user_cache_sdk)
        except Exception as e:
            # lambda succeeds if we're unable to cache the value
            CORVA_LOGGER.warning(f"Could not save data to cache. Details: {str(e)}.")

        return result

    return wrapper


def scheduled(
    func: Optional[Callable[[ScheduledEventT, Api, UserRedisSdk], Any]] = None,
    *,
    handler: Optional[logging.Handler] = None,
    merge_events: bool = False,
) -> Callable:
    """Runs scheduled app.

    Arguments:
        handler: logging handler to include in Corva logger.
        merge_events: if True - merge all incoming events into one before
          passing them to func
    """

    if func is None:
        return functools.partial(scheduled, handler=handler, merge_events=merge_events)

    @functools.wraps(func)
    @functools.partial(
        base_handler,
        raw_event_type=RawScheduledEvent,
        handler=handler,
        merge_events=merge_events,
    )
    def wrapper(
        event: RawScheduledEvent,
        api_key: str,
        aws_request_id: str,
        logging_ctx: LoggingContext,
        redis_client: redis.Redis,
    ) -> Any:
        logging_ctx.asset_id = event.asset_id
        logging_ctx.app_connection_id = event.app_connection_id

        api = Api(
            api_url=SETTINGS.API_ROOT_URL,
            data_api_url=SETTINGS.DATA_API_ROOT_URL,
            api_key=api_key,
            app_key=SETTINGS.APP_KEY,
            timeout=None,
            app_connection_id=event.app_connection_id,
        )

        hash_name = get_cache_key(
            provider=SETTINGS.PROVIDER,
            asset_id=event.asset_id,
            app_stream_id=event.app_stream_id,
            app_key=SETTINGS.APP_KEY,
            app_connection_id=event.app_connection_id,
        )

        user_cache_sdk = UserRedisSdk(
            hash_name=hash_name, redis_dsn=SETTINGS.CACHE_URL, redis_client=redis_client
        )
        internal_cache_sdk = InternalRedisSdk(
            hash_name=hash_name, redis_client=redis_client
        )

        app_event = event.scheduler_type.event.parse_obj(event)

        with LoggingContext(
            aws_request_id=aws_request_id,
            asset_id=event.asset_id,
            app_connection_id=event.app_connection_id,
            handler=CorvaLoggerHandler(
                max_message_size=SETTINGS.LOG_THRESHOLD_MESSAGE_SIZE,
                max_message_count=SETTINGS.LOG_THRESHOLD_MESSAGE_COUNT,
                logger=CORVA_LOGGER,
                placeholder=" ...",
            ),
            user_handler=handler,
            logger=CORVA_LOGGER,
        ):
            try:
                result = service.run_app(
                    has_secrets=event.has_secrets,
                    app_key=SETTINGS.APP_KEY,
                    api_sdk=CachingApiSdk(
                        api_sdk=CorvaApiSdk(api_adapter=api),
                        ttl=SETTINGS.SECRETS_CACHE_TTL,
                    ),
                    cache_sdk=internal_cache_sdk,
                    app=functools.partial(
                        cast(Callable[[ScheduledEvent, Api, UserRedisSdk], Any], func),
                        app_event,
                        api,
                        user_cache_sdk,
                    ),
                )
            except Exception:
                if isinstance(app_event, ScheduledNaturalTimeEvent):
                    set_schedule_as_completed(event=event, api=api)
                raise
            else:
                set_schedule_as_completed(event=event, api=api)

        return result

    return wrapper


def set_schedule_as_completed(event: RawScheduledEvent, api: Api) -> None:
    try:
        event.set_schedule_as_completed(api=api)
    except Exception as e:
        # lambda succeeds if we're unable to set completed status
        CORVA_LOGGER.warning(f"Could not set schedule as completed. Details: {str(e)}.")


def task(
    func: Optional[Callable[[TaskEvent, Api], Any]] = None,
    *,
    handler: Optional[logging.Handler] = None,
) -> Callable:
    """Runs task app.

    Arguments:
        handler: logging handler to include in Corva logger.
    """

    if func is None:
        return functools.partial(task, handler=handler)

    @functools.wraps(func)
    @functools.partial(base_handler, raw_event_type=RawTaskEvent, handler=handler)
    def wrapper(
        event: RawTaskEvent,
        api_key: str,
        aws_request_id: str,
        logging_ctx: LoggingContext,
        redis_client: redis.Redis,
    ) -> Any:
        status = TaskStatus.fail
        data: Dict[str, Union[dict, str]] = {"payload": {}}

        api = Api(
            api_url=SETTINGS.API_ROOT_URL,
            data_api_url=SETTINGS.DATA_API_ROOT_URL,
            api_key=api_key,
            app_key=SETTINGS.APP_KEY,
            timeout=None,
            app_connection_id=None,
        )

        try:
            app_event = event.get_task_event(api=api)

            logging_ctx.asset_id = app_event.asset_id

            with LoggingContext(
                aws_request_id=aws_request_id,
                asset_id=app_event.asset_id,
                app_connection_id=None,
                handler=CorvaLoggerHandler(
                    max_message_size=SETTINGS.LOG_THRESHOLD_MESSAGE_SIZE,
                    max_message_count=SETTINGS.LOG_THRESHOLD_MESSAGE_COUNT,
                    logger=CORVA_LOGGER,
                    placeholder=" ...",
                ),
                user_handler=handler,
                logger=CORVA_LOGGER,
            ):
                result = service.run_app(
                    has_secrets=event.has_secrets,
                    app_key=SETTINGS.APP_KEY,
                    api_sdk=CachingApiSdk(
                        api_sdk=CorvaApiSdk(api_adapter=api),
                        ttl=SETTINGS.SECRETS_CACHE_TTL,
                    ),
                    cache_sdk=FakeInternalCacheSdk(),
                    app=functools.partial(
                        cast(Callable[[TaskEvent, Api], Any], func), app_event, api
                    ),
                )

            if isinstance(result, dict):
                warnings.warn(
                    "Returning dict result from task app to get it stored in task "
                    "payload is deprecated and will be removed from corva in the next "
                    "major version. Update the payload explicitly in your app.",
                    FutureWarning,
                )

                data["payload"] = result

            status = TaskStatus.success

            return result

        except Exception as exc:
            CORVA_LOGGER.exception("Task app failed to execute.")
            data = {"fail_reason": str(exc)}
            raise

        finally:
            try:
                event.update_task_data(
                    api=api,
                    status=status,
                    data=data,
                ).raise_for_status()
            except Exception as e:
                # lambda succeeds if we're unable to update task data
                CORVA_LOGGER.warning(f"Could not update task data. Details: {str(e)}.")

    return wrapper


def partial_rerun_merge(
    func: Optional[
        Callable[[PartialRerunMergeEvent, Api, UserRedisSdk, UserRedisSdk], Any]
    ] = None,
    *,
    handler: Optional[logging.Handler] = None,
) -> Callable:
    """Runs partial merge app.

    Arguments:
        handler: logging handler to include in Corva logger.
    """

    if func is None:
        return functools.partial(partial_rerun_merge, handler=handler)

    @functools.wraps(func)
    def wrapper(
        event: RawPartialRerunMergeEvent,
        api_key: str,
        aws_request_id: str,
        logging_ctx: LoggingContext,
        redis_client: redis.Redis,
    ) -> Any:
        logging_ctx.asset_id = event.data.asset_id
        logging_ctx.app_connection_id = event.data.app_connection_id

        api = Api(
            api_url=SETTINGS.API_ROOT_URL,
            data_api_url=SETTINGS.DATA_API_ROOT_URL,
            api_key=api_key,
            app_key=SETTINGS.APP_KEY,
        )

        asset_cache_hash_name = get_cache_key(
            provider=SETTINGS.PROVIDER,
            asset_id=event.data.asset_id,
            app_stream_id=event.data.app_stream_id,
            app_key=SETTINGS.APP_KEY,
            app_connection_id=event.data.app_connection_id,
        )
        rerun_asset_cache_hash_name = get_cache_key(
            provider=SETTINGS.PROVIDER,
            asset_id=event.data.rerun_asset_id,
            app_stream_id=event.data.rerun_app_stream_id,
            app_key=SETTINGS.APP_KEY,
            app_connection_id=event.data.rerun_app_connection_id,
        )
        internal_cache_hash_name = get_cache_key(
            provider=SETTINGS.PROVIDER,
            asset_id=event.data.rerun_asset_id,
            app_stream_id=event.data.rerun_app_stream_id,
            app_key=SETTINGS.APP_KEY,
            app_connection_id=event.data.app_connection_id,
        )

        asset_cache = UserRedisSdk(
            hash_name=asset_cache_hash_name,
            redis_dsn=SETTINGS.CACHE_URL,
            redis_client=redis_client,
        )
        rerun_asset_cache = UserRedisSdk(
            hash_name=rerun_asset_cache_hash_name,
            redis_dsn=SETTINGS.CACHE_URL,
            redis_client=redis_client,
        )
        app_event = PartialRerunMergeEvent(
            **event.data.dict(), event_type=event.event_type
        )

        with LoggingContext(
            aws_request_id=aws_request_id,
            asset_id=event.data.asset_id,
            app_connection_id=event.data.app_connection_id,
            handler=CorvaLoggerHandler(
                max_message_size=SETTINGS.LOG_THRESHOLD_MESSAGE_SIZE,
                max_message_count=SETTINGS.LOG_THRESHOLD_MESSAGE_COUNT,
                logger=CORVA_LOGGER,
                placeholder=" ...",
            ),
            user_handler=handler,
            logger=CORVA_LOGGER,
        ):
            result = service.run_app(
                has_secrets=event.has_secrets,
                app_key=SETTINGS.APP_KEY,
                api_sdk=CachingApiSdk(
                    api_sdk=CorvaApiSdk(api_adapter=api),
                    ttl=SETTINGS.SECRETS_CACHE_TTL,
                ),
                cache_sdk=InternalRedisSdk(
                    hash_name=internal_cache_hash_name, redis_client=redis_client
                ),
                app=functools.partial(
                    cast(
                        Callable[
                            [PartialRerunMergeEvent, Api, UserRedisSdk, UserRedisSdk],
                            Any,
                        ],
                        func,
                    ),
                    app_event,
                    api,
                    asset_cache,
                    rerun_asset_cache,
                ),
            )

        return result

    HANDLERS[RawPartialRerunMergeEvent] = wrapper
    return wrapper


def _get_custom_event_type_by_raw_aws_event(
    aws_event: Any,
) -> Union[Tuple[Type[RawBaseEvent], Callable], Tuple[None, None]]:
    events = None
    # Here we do not know what schema will future custom events have,
    # so trying to parse all registered custom types.
    for event_type, handler in HANDLERS.items():
        with contextlib.suppress(pydantic.ValidationError):
            events = event_type.from_raw_event(aws_event)
        if events:
            return event_type, handler
    return None, None


def _merge_events(
    aws_event: Any,
    data_transformation_type: Type[RawBaseEvent],
) -> Any:
    """
    Merges incoming aws_events into one.
    Merge happens differently, depending on app type.
    Only "scheduled" and "stream" type of apps can be processed here.
    If somehow any other type is passed - raise an exception
    """
    if not isinstance(
        aws_event, list
    ):  # if aws_event is not a list - there is nothing to merge, so do nothing.
        return aws_event

    if data_transformation_type is RawScheduledEvent:
        # scheduled event
        if not isinstance(aws_event[0], dict):
            aws_event = list(itertools.chain(*aws_event))
        scheduler_type = aws_event[0]["scheduler_type"]
        if isinstance(scheduler_type, SchedulerType):
            scheduler_type = scheduler_type.value
        is_depth = scheduler_type == SchedulerType.data_depth_milestone.value
        event_start, event_end = (
            ("top_depth", "bottom_depth")
            if is_depth
            else ("schedule_start", "schedule_end")
        )
        min_event_start = min(e[event_start] for e in aws_event)
        max_event_end = max(
            (e[event_end] for e in aws_event if e.get(event_end) is not None),
            default=None,
        )
        aws_event[0][event_start] = min_event_start
        if max_event_end:
            aws_event[0][event_end] = max_event_end
        aws_event = aws_event[0]

    elif data_transformation_type is RawStreamEvent:
        # stream event
        for event in aws_event[1:]:
            aws_event[0]["records"].extend(event["records"])
        aws_event = [aws_event[0]]

    else:
        CORVA_LOGGER.warning(
            f"{data_transformation_type.__name__} does not support `merge event` "
            "parameter."
        )

    return aws_event



import datetime
import warnings
from typing import (
    Dict,
    List,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Union,
    cast,
    overload,
)

import fakeredis
import redis

from corva import cache_adapter


class UserCacheSdkProtocol(Protocol):
    def set(self, key: str, value: str, ttl: int = ...) -> None:
        ...

    def set_many(
        self, data: Sequence[Union[Tuple[str, str], Tuple[str, str, int]]]
    ) -> None:
        ...

    def get(self, key: str) -> Optional[str]:
        ...

    def get_many(self, keys: Sequence[str]) -> Dict[str, Optional[str]]:
        ...

    def get_all(self) -> Dict[str, str]:
        ...

    # TODO: remove asterisk in v2 - it was added for backward compatibility
    def delete(self, *, key: str) -> None:
        ...

    def delete_many(self, keys: Sequence[str]) -> None:
        ...

    def delete_all(self) -> None:
        ...


class UserRedisSdk:
    """User cache protocol implementation using Redis.

    As AWS Lambda is meant to be stateless, the apps need some mechanism to share the
    data between invokes. Redis provides an in-memory low latency storage for such data.
    """

    SIXTY_DAYS: int = int(datetime.timedelta(days=60).total_seconds())

    def __init__(
        self,
        hash_name: str,
        redis_dsn: str,
        use_fakes: bool = False,
        redis_client: Optional[redis.Redis] = None,
    ):
        use_lua_52 = False
        # use either provided redis client, or initialize "fake" client
        # (usually used for tests), or initialize real new client
        if use_fakes:
            redis_client = fakeredis.FakeRedis.from_url(
                url=redis_dsn, decode_responses=True
            )
            use_lua_52 = True
        elif redis_client is None:
            redis_client = redis.Redis.from_url(url=redis_dsn, decode_responses=True)

        self.cache_repo = cache_adapter.RedisRepository(
            hash_name=hash_name,
            client=cast(redis.Redis, redis_client),
            use_lua_52=use_lua_52,
        )
        self.old_cache_repo = cache_adapter.DeprecatedRedisAdapter(
            hash_name=hash_name, client=cast(redis.Redis, redis_client)
        )

    def set(self, key: str, value: str, ttl: int = SIXTY_DAYS) -> None:
        self.cache_repo.set(key=key, value=value, ttl=ttl)

    def set_many(
        self, data: Sequence[Union[Tuple[str, str], Tuple[str, str, int]]]
    ) -> None:
        prepared_data = [
            cast(Tuple[str, str, int], datum)
            if len(datum) == 3
            else (datum[0], datum[1], self.SIXTY_DAYS)
            for datum in data
        ]
        self.cache_repo.set_many(data=prepared_data)

    def get(self, key: str) -> Optional[str]:
        return self.cache_repo.get(key=key)

    def get_many(self, keys: Sequence[str]) -> Dict[str, Optional[str]]:
        return self.cache_repo.get_many(keys=keys)

    def get_all(self) -> Dict[str, str]:
        return self.cache_repo.get_all()

    @overload
    def delete(self, *, key: str) -> None:
        ...

    @overload
    def delete(self, keys: List[str], name: Optional[str] = ...) -> int:
        ...

    def delete(
        self,
        keys=None,
        name=None,
        *,
        key=None,
    ):
        if keys is not None:
            warnings.warn(
                "The `keys` and `name` kwargs of `delete` cache method are deprecated "
                "and will be removed from corva in the next major version. "
                "Use `key` kwarg instead.",
                FutureWarning,
            )
            return self.old_cache_repo.hdel(keys=keys, name=name)

        if key is not None:
            self.cache_repo.delete(key=key)
            return None

        raise TypeError("Bad arguments")

    def delete_many(self, keys: Sequence[str]) -> None:
        self.cache_repo.delete_many(keys=keys)

    @overload
    def delete_all(self) -> None:
        ...

    @overload
    def delete_all(self, *names: str):
        ...

    def delete_all(self, *names):
        if names:
            warnings.warn(
                "`delete_all` with `*names` parameter cache method is deprecated and"
                " will be removed from corva in the next major version. Use "
                "`delete_many` cache method instead.",
                FutureWarning,
            )
            return self.old_cache_repo.delete(*names)
        else:
            self.cache_repo.delete_all()

    def store(self, **kwargs):
        warnings.warn(
            "`store` cache method is deprecated and will be removed from corva in the"
            " next major version. Use `set` or `set_many` cache methods instead.",
            FutureWarning,
        )
        return self.old_cache_repo.hset(**kwargs)

    def load(self, **kwargs):
        warnings.warn(
            "`load` cache method is deprecated and will be removed from corva in the"
            " next major version. Use `get` or `get_many` cache methods instead.",
            FutureWarning,
        )
        return self.old_cache_repo.hget(**kwargs)

    def load_all(self, **kwargs):
        warnings.warn(
            "`load_all` cache method is deprecated and will be removed from corva in "
            "the next major version. Use `get_all` cache method instead.",
            FutureWarning,
        )
        return self.old_cache_repo.hgetall(**kwargs)

    def ttl(self, **kwargs):
        warnings.warn(
            "`ttl` cache method is deprecated and will be removed from corva in the"
            " next major version.",
            FutureWarning,
        )
        return self.old_cache_repo.ttl(**kwargs)

    def pttl(self, **kwargs):
        warnings.warn(
            "`pttl` cache method is deprecated and will be removed from corva in the"
            " next major version.",
            FutureWarning,
        )
        return self.old_cache_repo.pttl(**kwargs)

    def exists(self, *names):
        warnings.warn(
            "`exists` cache method is deprecated and will be removed from corva in the"
            " next major version.",
            FutureWarning,
        )
        return self.old_cache_repo.exists(*names)


class InternalCacheSdkProtocol(Protocol):
    def vacuum(self, delete_count: int) -> None:
        ...


class InternalRedisSdk:
    def __init__(
        self,
        hash_name: str,
        redis_client: redis.Redis,
    ):
        self.cache_repo = cache_adapter.RedisRepository(
            hash_name=hash_name,
            client=redis_client,
        )

    def vacuum(self, delete_count: int) -> None:
        self.cache_repo.vacuum(delete_count=delete_count)


class FakeInternalCacheSdk:
    def __init__(self):
        self.vacuum_called = False

    def vacuum(self, delete_count: int) -> None:
        self.vacuum_called = True



import functools
import inspect
from types import SimpleNamespace
from typing import Any, Callable, ClassVar, Dict, Optional, Union

from corva.api import Api
from corva.configuration import SETTINGS
from corva.models.merge.merge import PartialRerunMergeEvent
from corva.models.scheduled.scheduled import ScheduledEvent
from corva.models.stream.stream import StreamEvent
from corva.models.task import TaskEvent
from corva.service import service
from corva.service.api_sdk import FakeApiSdk
from corva.service.cache_sdk import FakeInternalCacheSdk, UserRedisSdk


class TestClient:
    """Interface for testing Corva apps.

    Attributes:
        _context: AWS Lambda context expected by Corva.
        _api: Api instance.
    """

    _context: ClassVar[SimpleNamespace] = SimpleNamespace(
        aws_request_id="qwerty", client_context=SimpleNamespace(env={"API_KEY": "123"})
    )
    _api: ClassVar[Api] = Api(
        api_url=SETTINGS.API_ROOT_URL,
        data_api_url=SETTINGS.DATA_API_ROOT_URL,
        api_key=_context.client_context.env["API_KEY"],
        app_key=SETTINGS.APP_KEY,
    )

    @staticmethod
    def run(
        fn: Callable,
        event: Union[TaskEvent, StreamEvent, ScheduledEvent, PartialRerunMergeEvent],
        *,
        secrets: Optional[Dict[str, str]] = None,
        cache: Optional[UserRedisSdk] = None
    ) -> Any:

        app: Callable[[], Any]

        if isinstance(event, TaskEvent):
            app = functools.partial(inspect.unwrap(fn), event, TestClient._api)

        if isinstance(event, (ScheduledEvent, StreamEvent)):
            if not cache:
                cache = UserRedisSdk(
                    hash_name="hash_name",
                    redis_dsn=SETTINGS.CACHE_URL,
                    use_fakes=True,
                )

            app = functools.partial(
                inspect.unwrap(fn),
                event,
                TestClient._api,
                cache,
            )
        if isinstance(event, PartialRerunMergeEvent):
            app = functools.partial(
                inspect.unwrap(fn),
                event,
                TestClient._api,
                UserRedisSdk(
                    hash_name="asset_cache",
                    redis_dsn=SETTINGS.CACHE_URL,
                    use_fakes=True,
                ),
                UserRedisSdk(
                    hash_name="rerun_asset_cache",
                    redis_dsn=SETTINGS.CACHE_URL,
                    use_fakes=True,
                ),
            )

        return service.run_app(
            has_secrets=True,
            app_key="test_app_key",
            api_sdk=FakeApiSdk(secrets={"test_app_key": secrets or {}}),
            cache_sdk=FakeInternalCacheSdk(),
            app=app,
        )



import enum


class LogType(enum.Enum):
    time = 'time'
    depth = 'depth'

    @property
    def raw_event(self):
        from corva.models.stream.raw import RawStreamDepthEvent, RawStreamTimeEvent

        mapping = {
            type(self).time: RawStreamTimeEvent,
            type(self).depth: RawStreamDepthEvent,
        }
        return mapping[self]

    @property
    def event(self):
        from corva.models.stream.stream import StreamDepthEvent, StreamTimeEvent

        mapping = {type(self).time: StreamTimeEvent, type(self).depth: StreamDepthEvent}
        return mapping[self]



from typing import Any, Callable
from unittest import mock

from corva import shared
from corva.service.api_sdk import ApiSdkProtocol
from corva.service.cache_sdk import InternalCacheSdkProtocol


def run_app(
    has_secrets: bool,
    app_key: str,
    api_sdk: ApiSdkProtocol,
    cache_sdk: InternalCacheSdkProtocol,
    app: Callable[[], Any],
    vacuum_delete_count: int = 3,
) -> Any:
    cache_sdk.vacuum(delete_count=vacuum_delete_count)

    secrets = api_sdk.get_secrets(app_key=app_key) if has_secrets else {}

    with mock.patch.dict(shared.SECRETS, values=secrets, clear=True):
        result = app()

    return result



import contextlib
import logging
import logging.config
import sys
import time
from typing import Optional

from corva.configuration import SETTINGS

logging.Formatter.converter = time.gmtime  # log time as UTC

CORVA_LOGGER = logging.getLogger('corva')
CORVA_LOGGER.setLevel(SETTINGS.LOG_LEVEL)

# unset to pass messages to ancestor loggers, including OTel Log Sending handler
# see https://github.com/corva-ai/otel/pull/37
# see https://corvaqa.atlassian.net/browse/EE-31
# CORVA_LOGGER.propagate = False


def get_formatter(
    aws_request_id: bool, asset_id: bool, app_connection_id: bool
) -> logging.Formatter:
    return logging.Formatter(
        f'%(asctime)s.%(msecs)03dZ '
        f'{"%(aws_request_id)s " if aws_request_id else ""}'
        f'%(levelname)s '
        f'{"ASSET=%(asset_id)s " if asset_id else ""}'
        f'{"AC=%(app_connection_id)s " if app_connection_id else ""}'
        f'| %(message)s',
        '%Y-%m-%dT%H:%M:%S',
    )


class CorvaLoggerFilter(logging.Filter):
    """Injects fields into logging.LogRecord instance for usage in logging.Formatter."""

    def __init__(
        self,
        aws_request_id: str,
        asset_id: Optional[int],
        app_connection_id: Optional[int],
    ):
        logging.Filter.__init__(self)

        self.aws_request_id = aws_request_id
        self.asset_id = asset_id
        self.app_connection_id = app_connection_id

    def filter(self, record):
        record.aws_request_id = self.aws_request_id
        record.asset_id = self.asset_id
        record.app_connection_id = self.app_connection_id

        return True


class CorvaLoggerHandler(logging.Handler):
    """Logging handler with constraints.

    The handler logs to sys.stdout and has following functionality:
        1. Truncates the message to not exceed max message size.
        2. Disables the logging after reaching max message count and
            logs corresponding warning.

    Args:
        max_message_size: Maximum allowed message size in bytes.
            Messages that exceed this limit get truncated.
        max_message_count: Maximum allowed number of logged messages.
            After reaching this limit logging gets disabled.
        logger: Logger that is used to log the warning about reaching max
            message count.
        placeholder: String that will appear at the end of the message
            if it has been truncated.
    """

    TERMINATOR = '\n'

    def __init__(
        self,
        max_message_size: int,
        max_message_count: int,
        logger: logging.Logger,
        placeholder: str,
    ):
        logging.Handler.__init__(self)

        self.max_message_size = max_message_size
        self.max_message_count = max_message_count
        self.logger = logger
        self.placeholder = f'{placeholder}{self.TERMINATOR}'

        self.logging_warning = False
        # one extra message to log the warning
        self.residue_message_count = self.max_message_count + 1

    def emit(self, record: logging.LogRecord) -> None:
        if self.residue_message_count == 0:
            return

        if self.residue_message_count > 1 or self.logging_warning:
            self.log(message=self.format(record))
            self.residue_message_count -= 1

        if self.residue_message_count == 1:
            self.logging_warning = True
            self.logger.warning(
                f'Disabling the logging as maximum number of logged messages '
                f'was reached: {self.max_message_count}.'
            )

    def format(self, record: logging.LogRecord) -> str:
        message = super().format(record)

        # https://github.com/debug-js/debug/issues/296#issuecomment-289595923
        # For CloudWatch `\n` means end of whole log and `\r` means end of line in
        # multiline logs.
        # Replace `\n` for `\r` for logs to display correctly in CloudWatch.
        message = message.replace('\n', '\r')
        message = f'{message}{self.TERMINATOR}'

        extra_chars_count = len(message) - self.max_message_size

        if extra_chars_count <= 0:
            # no need to truncate the message
            return message

        message_end_idx = len(message) - (extra_chars_count + len(self.placeholder))

        if message_end_idx <= 0:
            return ''

        message = message[:message_end_idx] + self.placeholder

        return message

    def log(self, message: str) -> None:
        sys.stdout.write(message)


class LoggingContext(contextlib.ContextDecorator):
    """Context manager to configure logger to use specified handlers.

    User handler does not get any modifications.

    Handler gets following modifications:
        - Added CorvaLoggerFilter to filters.
        - Set log level from settings.
        - Set Corva formatter.

    Context allows changing filter's fields dynamically. It will update logging
    formatter as needed.

    Logger gets its old handlers back at the end of the context.

    Attributes:
        filter: Logging filter that gets set in the handler.
        handler: Logging handler that gets modified and set in the logger.
        user_handler: Logging handler that gets set in the logger without modifications.
        logger: Logger to configure.
        old_handlers: Logger's own handlers before the update.
    """

    def __init__(
        self,
        aws_request_id: str,
        asset_id: Optional[int],
        app_connection_id: Optional[int],
        handler: logging.Handler,
        user_handler: Optional[logging.Handler],
        logger: logging.Logger,
    ):
        self.filter = CorvaLoggerFilter(
            aws_request_id=aws_request_id,
            asset_id=asset_id,
            app_connection_id=app_connection_id,
        )

        self.handler = handler
        self.handler.setLevel(SETTINGS.LOG_LEVEL)
        self.handler.addFilter(self.filter)
        self.set_formatter()

        self.user_handler = user_handler
        self.logger = logger

    @property
    def asset_id(self) -> Optional[int]:
        """Asset id used in the logging filter."""

        return self.filter.asset_id

    @asset_id.setter
    def asset_id(self, value: Optional[int]) -> None:
        self.filter.asset_id = value
        self.set_formatter()

    @property
    def app_connection_id(self) -> Optional[int]:
        """App connection id used in the logging filter."""

        return self.filter.app_connection_id

    @app_connection_id.setter
    def app_connection_id(self, value: Optional[int]) -> None:
        self.filter.app_connection_id = value
        self.set_formatter()

    def set_formatter(self):
        self.handler.setFormatter(
            get_formatter(
                aws_request_id=True,
                asset_id=self.asset_id is not None,
                app_connection_id=self.app_connection_id is not None,
            )
        )

    def __enter__(self):
        self.old_handlers = self.logger.handlers
        self.logger.handlers = (
            [self.handler, self.user_handler] if self.user_handler else [self.handler]
        )

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.logger.handlers = self.old_handlers

        return False  # exception will be propagated



from __future__ import annotations

import abc
from typing import Any, Sequence

import pydantic


class CorvaBaseEvent(pydantic.BaseModel):
    class Config:
        extra = pydantic.Extra.allow
        allow_mutation = False


class RawBaseEvent(abc.ABC):
    @staticmethod
    @abc.abstractmethod
    def from_raw_event(event: Any) -> Sequence[RawBaseEvent]:
        pass



from typing import TYPE_CHECKING, Optional, Sequence

import pydantic

from corva.models.base import CorvaBaseEvent
from corva.models.rerun import RerunDepth, RerunTime


class StreamTimeRecord(CorvaBaseEvent):
    """Stream time record data.

    Attributes:
        timestamp: Unix timestamp.
        data: record data.
        metadata: record metadata.
    """

    timestamp: int
    data: dict = {}
    metadata: dict = {}


class StreamDepthRecord(CorvaBaseEvent):
    """Stream depth record data.

    Attributes:
        measured_depth: measured depth (ft).
        data: record data.
        metadata: record metadata.
    """

    measured_depth: float
    data: dict = {}
    metadata: dict = {}


if TYPE_CHECKING:
    RecordsTime = Sequence[StreamTimeRecord]
    RecordsDepth = Sequence[StreamDepthRecord]
else:
    RecordsTime = pydantic.conlist(StreamTimeRecord, min_items=1)
    RecordsDepth = pydantic.conlist(StreamDepthRecord, min_items=1)


class StreamEvent(CorvaBaseEvent):
    """Base stream event data."""


class StreamTimeEvent(StreamEvent):
    """Stream time event data.

    Attributes:
        asset_id: asset id.
        company_id: company id.
        records: data records.
        rerun: rerun metadata.
    """

    asset_id: int
    company_id: int
    records: RecordsTime
    rerun: Optional[RerunTime] = None


class StreamDepthEvent(StreamEvent):
    """Stream depth event data.

    Attributes:
        asset_id: asset id.
        company_id: company id.
        records: data records.
        rerun: rerun metadata.
    """

    asset_id: int
    company_id: int
    records: RecordsDepth
    # TODO: remove `Optional` in v2 as it was added for backward compatibility.
    log_identifier: Optional[str] = None
    rerun: Optional[RerunDepth] = None



from __future__ import annotations

from typing import Any, Dict, List, Optional

import pydantic

from corva.models.base import CorvaBaseEvent, RawBaseEvent
from corva.models.merge.enums import EventType, RerunMode, SourceType


class RawPartialMergeEventData(pydantic.BaseModel):
    partial_well_rerun_id: Optional[int]
    rerun_mode: Optional[RerunMode]
    start: Optional[int]
    end: Optional[int]
    asset_id: int
    rerun_asset_id: int
    app_stream_id: int
    rerun_app_stream_id: int
    version: int  # Partial re-run version
    app_id: Optional[int]
    app_key: Optional[str]
    app_connection_id: int
    rerun_app_connection_id: int
    source_type: Optional[SourceType]
    log_type: Optional[str]
    run_until: Optional[int]

    class Config:
        extra = "allow"


class RawPartialRerunMergeEvent(CorvaBaseEvent, RawBaseEvent):
    event_type: EventType
    data: RawPartialMergeEventData
    has_secrets: bool = False

    @staticmethod
    def from_raw_event(event: Dict[str, Any]) -> List[RawPartialRerunMergeEvent]:
        return [pydantic.parse_obj_as(RawPartialRerunMergeEvent, event)]



from corva.models.base import CorvaBaseEvent
from corva.models.stream.log_type import LogType


class InitialMetadata(CorvaBaseEvent):
    log_type: LogType


class InitialStreamEvent(CorvaBaseEvent):
    """Stores the most essential data, that is parsed first."""

    metadata: InitialMetadata



from typing import Optional

import pydantic

from corva.models.base import CorvaBaseEvent
from corva.models.rerun import RerunDepth, RerunTime


class ScheduledEvent(CorvaBaseEvent):
    """Base class for scheduled event data."""


class ScheduledDataTimeEvent(ScheduledEvent):
    """Data time scheduled event data.

    Data time scheduled apps are run when there is a time span worth of new data.
    The time span is specified by user (e.g., 1 minute, 5 minutes, etc.). The event
    contains time range (start and end times) that there is a new data for. The app
    may then fetch and use that data.

    Attributes:
        asset_id: asset id.
        company_id: company id.
        start_time: left bound of the time range, covered by this event.
            Use inclusively.
        end_time: right bound of the time range, covered by this event.
            Use inclusively.
        rerun: rerun metadata.
    """

    asset_id: int
    company_id: int
    start_time: int
    end_time: int = pydantic.Field(..., alias='schedule_start')
    rerun: Optional[RerunTime] = None

    class Config:
        allow_population_by_field_name = True


class ScheduledDepthEvent(ScheduledEvent):
    """Depth scheduled event data.

    Depth scheduled apps are run when there is a distance worth of new data.
    The distance is specified by user (e.g., 5 ft., 10 ft., etc.). The event contains
    depth range (top and bottom depths) that there is a new data for. The app
    may then fetch and use that data.

    Attributes:
        asset_id: asset id.
        company_id: company id.
        top_depth: start depth in ft., covered by this event. Use exclusively.
        bottom_depth: end depth in ft., covered by this event. Use inclusively.
        log_identifier: app stream log identifier. Used to scope data by stream,
            asset might be connected to multiple depth based logs.
        interval: distance between two schedule triggers.
        rerun: rerun metadata.
    """

    asset_id: int
    company_id: int
    top_depth: float
    bottom_depth: float
    log_identifier: str
    interval: float
    rerun: Optional[RerunDepth] = None


class ScheduledNaturalTimeEvent(ScheduledEvent):
    """Natural time scheduled event data.

    Natural time scheduled apps are run with time frequency set up by user based on
    the actual time instead of data time (e.g., every 1 minute, every 5 minutes, etc.).

    Attributes:
        asset_id: asset id.
        company_id: company id.
        schedule_start: schedule trigger time.
        interval: time between two schedule triggers.
        rerun: rerun metadata.
    """

    asset_id: int
    company_id: int
    schedule_start: int
    interval: float
    rerun: Optional[RerunTime] = None



from __future__ import annotations

import contextlib
from typing import Any, ContextManager
from unittest import mock

import pydantic


class AwsEventWithClientContext(pydantic.BaseModel):
    client_context: dict


class Env(pydantic.BaseModel):
    API_KEY: str


class ClientContext(pydantic.BaseModel):
    class Config:
        orm_mode = True

    env: Env


class CorvaContext(pydantic.BaseModel):
    """AWS context, expected by Corva."""

    class Config:
        orm_mode = True

    aws_request_id: str
    client_context: ClientContext

    @property
    def api_key(self) -> str:
        return self.client_context.env.API_KEY

    @classmethod
    def from_aws(cls, aws_event: Any, aws_context: Any) -> CorvaContext:
        parse_ctx: ContextManager = contextlib.nullcontext()

        if aws_context.client_context is None:
            parse_ctx = mock.patch.object(
                aws_context,
                'client_context',
                AwsEventWithClientContext.parse_obj(aws_event).client_context,
            )

        with parse_ctx:
            return CorvaContext.from_orm(aws_context)



import enum
from typing import Type

import pydantic


class SchedulerType(enum.Enum):
    natural_time = 1
    data_time = 2
    data_depth_milestone = 4

    @property
    def raw_event(self):
        from corva.models.scheduled.raw import (
            RawScheduledDataTimeEvent,
            RawScheduledDepthEvent,
            RawScheduledNaturalTimeEvent,
        )

        mapping = {
            type(self).natural_time: RawScheduledNaturalTimeEvent,
            type(self).data_time: RawScheduledDataTimeEvent,
            type(self).data_depth_milestone: RawScheduledDepthEvent,
        }
        return mapping[self]

    @property
    def event(self) -> Type[pydantic.BaseModel]:
        from corva.models.scheduled.scheduled import (
            ScheduledDataTimeEvent,
            ScheduledDepthEvent,
            ScheduledNaturalTimeEvent,
        )

        mapping = {
            type(self).natural_time: ScheduledNaturalTimeEvent,
            type(self).data_time: ScheduledDataTimeEvent,
            type(self).data_depth_milestone: ScheduledDepthEvent,
        }
        return mapping[self]



from __future__ import annotations

import abc
import copy
from typing import TYPE_CHECKING, ClassVar, List, Optional, Sequence, Union

import pydantic

from corva.configuration import SETTINGS
from corva.models.base import CorvaBaseEvent, RawBaseEvent
from corva.models.rerun import RerunDepth, RerunTime
from corva.models.stream.initial import InitialStreamEvent
from corva.models.stream.log_type import LogType
from corva.service.cache_sdk import UserCacheSdkProtocol


class RawBaseRecord(CorvaBaseEvent, abc.ABC):
    asset_id: int
    company_id: int
    collection: str

    data: dict = {}
    metadata: dict = {}

    @property
    @abc.abstractmethod
    def record_value(self) -> Union[int, float]:
        pass


class RawTimeRecord(RawBaseRecord):
    timestamp: int
    measured_depth: Optional[float] = None

    @property
    def record_value(self) -> int:
        return self.timestamp


class RawDepthRecord(RawBaseRecord):
    timestamp: Optional[int] = None
    measured_depth: float

    @property
    def record_value(self) -> float:
        return self.measured_depth


class RawAppMetadata(CorvaBaseEvent):
    app_connection_id: int
    has_secrets: bool = False


class RawMetadata(CorvaBaseEvent):
    app_stream_id: int
    apps: pydantic.create_model(  # type: ignore
        "Apps",  # noqa: F821
        **{
            SETTINGS.APP_KEY: (
                RawAppMetadata,
                ...,
            )
        },
    )
    log_type: LogType
    # TODO: remove `Optional` in v2 as it was added for backward compatibility.
    log_identifier: Optional[str] = None


if TYPE_CHECKING:
    RecordsBase = Sequence[RawBaseRecord]
    RecordsTime = Sequence[RawTimeRecord]
    RecordsDepth = Sequence[RawDepthRecord]
else:
    RecordsBase = pydantic.conlist(RawBaseRecord, min_items=1)
    RecordsTime = pydantic.conlist(RawTimeRecord, min_items=0)
    RecordsDepth = pydantic.conlist(RawDepthRecord, min_items=1)


class RawStreamEvent(CorvaBaseEvent, RawBaseEvent):
    records: RecordsBase
    metadata: RawMetadata
    asset_id: int = None  # type: ignore
    company_id: int = None  # type: ignore

    # private attributes
    _max_record_value_cache_key: ClassVar[str]

    @property
    def app_connection_id(self) -> int:
        return getattr(self.metadata.apps, SETTINGS.APP_KEY).app_connection_id

    @property
    def has_secrets(self) -> bool:
        return getattr(self.metadata.apps, SETTINGS.APP_KEY).has_secrets

    @property
    def app_stream_id(self) -> int:
        return self.metadata.app_stream_id

    @property
    def is_completed(self) -> bool:
        """Indicates whether there is a completed record.

        There can only be 1 completed record always located at the end of the list.
        """

        if not self.records:
            return False

        return self.records[-1].collection == "wits.completed"

    @property
    def max_record_value(self) -> Union[int, float]:
        return max(record.record_value for record in self.records)

    @staticmethod
    def from_raw_event(event: List[dict]) -> List[RawStreamEvent]:
        initial_events: List[InitialStreamEvent] = pydantic.parse_obj_as(
            List[InitialStreamEvent], event
        )

        result = [
            initial_event.metadata.log_type.raw_event.parse_obj(sub_event)
            for initial_event, sub_event in zip(initial_events, event)
        ]

        return result

    def get_cached_max_record_value(
        self, cache: UserCacheSdkProtocol
    ) -> Optional[float]:
        result = cache.get(key=self._max_record_value_cache_key)

        if result is None:
            return result

        return float(result)

    def set_cached_max_record_value(self, cache: UserCacheSdkProtocol) -> None:
        cache.set(
            key=self._max_record_value_cache_key, value=str(self.max_record_value)
        )

    def filter_records(
        self,
        old_max_record_value: Optional[float],
    ) -> List[RawBaseRecord]:
        new_records = copy.deepcopy(self.records)

        if self.is_completed:
            new_records = new_records[:-1]  # remove "completed" record

        if old_max_record_value is None:
            return list(new_records)

        values = [record.record_value for record in new_records]

        new_records = [
            record
            for record, value in zip(new_records, values)
            if value > old_max_record_value
        ]

        return new_records

    @pydantic.root_validator(pre=False, skip_on_failure=True)
    def set_asset_id(cls, values: dict) -> dict:
        """Calculates asset_id field."""

        records: List[RawBaseRecord] = values["records"]

        if records:
            values["asset_id"] = int(records[0].asset_id)

        return values

    @pydantic.root_validator(pre=False, skip_on_failure=True)
    def set_company_id(cls, values: dict) -> dict:
        """Calculates company_id field."""

        records: List[RawBaseRecord] = values["records"]

        if records:
            values["company_id"] = int(records[0].company_id)

        return values

    @pydantic.validator("records", pre=True)
    def validate_records(cls, v):
        if isinstance(v, List):
            return [
                record
                for record in v
                if (
                    (isinstance(record, dict) and record.get("data") is not None)
                    or (hasattr(record, "data") and record.data is not None)
                )
            ]
        return v


class RawStreamTimeEvent(RawStreamEvent):
    records: RecordsTime
    rerun: Optional[RerunTime] = None
    _max_record_value_cache_key: ClassVar[str] = "last_processed_timestamp"


class RawStreamDepthEvent(RawStreamEvent):
    records: RecordsDepth
    rerun: Optional[RerunDepth] = None
    _max_record_value_cache_key: ClassVar[str] = "last_processed_depth"
    log_identifier: str = None  # type: ignore

    @pydantic.root_validator(pre=False, skip_on_failure=True)
    def set_log_identifier(cls, values: dict) -> dict:
        """Calculates log_identifier field."""

        metadata: RawMetadata = values["metadata"]

        values["log_identifier"] = metadata.log_identifier

        return values



import json
import posixpath
import re
from http import HTTPStatus
from typing import List, Optional, Sequence, Union

import requests
from tenacity import (
    RetryError,
    retry,
    retry_if_result,
    stop_after_attempt,
    wait_random_exponential,
)


class Api:
    """Provides a convenient way to access the Corva Platform API and Corva Data API.

    Api wraps the Python `requests` library and adds automatic authorization,
    convenient URL usage and reasonable timeouts to API requests.
    """

    TIMEOUT_LIMITS = (3, 30)  # seconds
    DEFAULT_MAX_RETRIES = int(0)

    def __init__(
        self,
        *,
        api_url: str,
        data_api_url: str,
        api_key: str,
        app_key: str,
        timeout: Optional[int] = None,
        app_connection_id: Optional[int] = None,
    ):
        self.api_url = api_url
        self.data_api_url = data_api_url
        self.api_key = api_key
        self.app_key = app_key
        self.app_connection_id = app_connection_id
        self.timeout = timeout or self.TIMEOUT_LIMITS[1]
        self._max_retries = self.DEFAULT_MAX_RETRIES

    @property
    def default_headers(self):
        return {
            "Authorization": f"API {self.api_key}",
            "X-Corva-App": self.app_key,
        }

    @property
    def max_retries(self) -> int:
        return self._max_retries

    @max_retries.setter
    def max_retries(self, value: int):
        if not (0 <= value <= 10):
            raise ValueError("Values between 0 and 10 are allowed")
        self._max_retries = value

    def get(self, path: str, **kwargs):
        return self._request("GET", path, **kwargs)

    def post(self, path: str, **kwargs):
        return self._request("POST", path, **kwargs)

    def patch(self, path: str, **kwargs):
        return self._request("PATCH", path, **kwargs)

    def put(self, path: str, **kwargs):
        return self._request("PUT", path, **kwargs)

    def delete(self, path: str, **kwargs):
        return self._request("DELETE", path, **kwargs)

    def _get_url(self, path: str):
        """Builds complete url.

        Args:
          path: either complete or only path part of the HTTP URL.

        Returns:
          1 path param, if path is a complete url.
          2 data api url, if path contains data api url pattern.
          3 corva api url, if above points are False.
        """

        if path.startswith("http"):
            return path

        path = path.lstrip(
            "/"
        )  # delete leading forward slash for posixpath.join to work correctly

        # search text like api/v1 or api/v10 in path
        if bool(re.search(r"api/v\d+", path)):
            return posixpath.join(self.data_api_url, path)

        return posixpath.join(self.api_url, path)

    @staticmethod
    def _execute_request(
        method: str,
        url: str,
        params: Optional[dict],
        data: Optional[dict],
        headers: Optional[dict] = None,
        timeout: Optional[int] = None,
    ):
        """Executes the request.

        Args:
            method: HTTP method.
            path: url to call.
            data: request body, that will be casted to json.
            params: url query string params.
            headers: additional headers to include in request.
            timeout: custom request timeout in seconds.

        Returns:
            requests.Response instance.
        """
        return requests.request(
            method=method,
            url=url,
            params=params,
            json=data,
            headers=headers,
            timeout=timeout,
        )

    def _request(
        self,
        method: str,
        path: str,
        *,
        data: Optional[dict] = None,
        params: Optional[dict] = None,
        headers: Optional[dict] = None,
        timeout: Optional[int] = None,
    ) -> requests.Response:
        """Prepares HTTP request.

        Args:
          method: HTTP method.
          path: url to call.
          data: request body, that will be casted to json.
          params: url query string params.
          headers: additional headers to include in request.
          timeout: custom request timeout in seconds.

        Returns:
          requests.Response instance.
        """
        retryable_status_codes = [
            HTTPStatus.TOO_MANY_REQUESTS,  # 428
            HTTPStatus.INTERNAL_SERVER_ERROR,  # 500
            HTTPStatus.BAD_GATEWAY,  # 502
            HTTPStatus.SERVICE_UNAVAILABLE,  # 503
            HTTPStatus.GATEWAY_TIMEOUT,  # 504
        ]

        timeout = timeout or self.timeout
        self._validate_timeout(timeout)

        url = self._get_url(path)

        headers = {
            **self.default_headers,
            **(headers or {}),
        }

        if self.max_retries > 0:
            retry_decorator = retry(
                stop=stop_after_attempt(self.max_retries),
                wait=wait_random_exponential(multiplier=0.25, max=10),
                retry=retry_if_result(
                    lambda r: r.status_code in retryable_status_codes
                ),
            )
            retrying_request = retry_decorator(self._execute_request)
            try:
                response = retrying_request(
                    method=method,
                    url=url,
                    params=params,
                    data=data,
                    headers=headers,
                    timeout=timeout,
                )
            except RetryError as e:
                if not e.last_attempt.failed:
                    response = e.last_attempt.result()
                else:
                    raise
        else:
            response = self._execute_request(
                method=method,
                url=url,
                params=params,
                data=data,
                headers=headers,
                timeout=timeout,
            )

        return response

    def _validate_timeout(self, timeout: int) -> None:
        if self.TIMEOUT_LIMITS[0] > timeout or self.TIMEOUT_LIMITS[1] < timeout:
            raise ValueError(
                f"Timeout must be between {self.TIMEOUT_LIMITS[0]} and "
                f"{self.TIMEOUT_LIMITS[1]} seconds."
            )

    def get_dataset(
        self,
        provider: str,
        dataset: str,
        *,
        query: dict,
        sort: dict,
        limit: int,
        skip: int = 0,
        fields: Optional[str] = None,
    ) -> List[dict]:
        """Fetches data from the endpoint '/api/v1/data/{provider}/{dataset}/'.

        Args:
          provider: company name, that owns the dataset.
          dataset: dataset name.
          query: search conditions. Example: {"asset_id": 123} - will fetch data
            for asset with id 123.
          sort: sort conditions. Example: {"timestamp": 1} - will sort data
            in ascending order by timestamp.
          limit: number of data points to fecth.
            Recommendation for setting the limit:
              1. The bigger ↑ each data point is - the smaller ↓ the limit;
              2. The smaller ↓ each data point is - the bigger ↑ the limit.
          skip: exclude from a response the first N items of a dataset.
          fields: comma separated list of fields to return. Example: "_id,data".

        Raises:
          requests.HTTPError: if request was unsuccessful.

        Returns:
          Data from dataset.
        """

        response = self.get(
            f"/api/v1/data/{provider}/{dataset}/",
            params={
                "query": json.dumps(query),
                "sort": json.dumps(sort),
                "fields": fields,
                "limit": limit,
                "skip": skip,
            },
        )
        response.raise_for_status()

        data = list(response.json())

        return data

    def produce_messages(self, data: Sequence[dict]) -> None:
        """Posts data to the endpoint '/api/v1/message_producer/'.

        Args:
          data: messages to post.
            Message examples:
            - time message [{"timestamp": 1}];
            - depth message [{"measured_depth": 1.0}].

        Raises:
          requests.HTTPError: if request was unsuccessful.
        """

        response = self.post(
            "/api/v1/message_producer/",
            data={"app_connection_id": self.app_connection_id, "data": data},
        )
        response.raise_for_status()

    def insert_data(
        self,
        provider: str,
        dataset: str,
        data: Sequence[dict],
        *,
        produce: bool = False,
    ) -> dict:
        """Posts data to the endpoint '/api/v1/data/{provider}/{dataset}/'.

        Args:
          provider: company name, that owns the dataset.
          dataset: dataset name.
          data: documents to insert.
          produce: whether to send data to message producer.

        Raises:
          requests.HTTPError: if request was unsuccessful.

        Returns: response dict.
        """

        body: Union[dict, List[dict]]  # make mypy happy

        if produce:
            body = {
                "data": list(data),
                "producer": {"app_connection_id": self.app_connection_id},
            }
        else:
            body = list(data)

        response = self.post(f"/api/v1/data/{provider}/{dataset}/", data=body)
        response.raise_for_status()

        return response.json()



from typing import Optional

from corva.models.base import CorvaBaseEvent
from corva.models.merge.enums import EventType, RerunMode, SourceType


class PartialRerunMergeEvent(CorvaBaseEvent):
    """Partial Merge event data.

    Attributes:
        event_type: EventType.PARTIAL_WELL_RERUN_MERGE
        partial_well_rerun_id: partial well rerun id
        rerun_mode: rerun mode
        start: start
        end: end
        asset_id: asset id
        rerun_asset_id: rerun asset id
        app_stream_id: app stream id
        rerun_app_stream_id: rerun app stream id
        version: version
        app_id: app id
        app_key: app key
        app_connection_id: app connection id
        rerun_app_connection_id: rerun app connection id
        source_type: source type
        log_type: log type
        run_until: run until
    """

    event_type: Optional[EventType]
    partial_well_rerun_id: Optional[int]
    rerun_mode: Optional[RerunMode]
    start: Optional[int]
    end: Optional[int]
    asset_id: int
    rerun_asset_id: int
    app_stream_id: int
    rerun_app_stream_id: int
    version: Optional[int]
    app_id: Optional[int]
    app_key: Optional[str]
    app_connection_id: int
    rerun_app_connection_id: int
    source_type: Optional[SourceType]
    log_type: Optional[str]
    run_until: Optional[int]

    class Config:
        extra = "allow"



from __future__ import annotations

import enum
from typing import List

import pydantic
import requests

from corva.api import Api
from corva.models.base import CorvaBaseEvent, RawBaseEvent


class TaskStatus(enum.Enum):
    fail = 'fail'
    success = 'success'


class TaskEvent(CorvaBaseEvent):
    """Task event data.

    Attributes:
        asset_id: asset id.
        company_id: company id.
        properties: custom task data.
    """

    asset_id: int
    company_id: int
    properties: dict = {}


class RawTaskEvent(CorvaBaseEvent, RawBaseEvent):
    task_id: str
    has_secrets: bool = False
    version: int = pydantic.Field(..., le=2, ge=2)  # only utils API v2 supported

    @staticmethod
    def from_raw_event(event: dict) -> List[RawTaskEvent]:
        return [pydantic.parse_obj_as(RawTaskEvent, event)]

    def get_task_event(self, api: Api) -> TaskEvent:
        response = api.get(path=f'v2/tasks/{self.task_id}')
        response.raise_for_status()

        return TaskEvent(**response.json())

    def update_task_data(
        self, api: Api, status: TaskStatus, data: dict
    ) -> requests.Response:
        """Updates the task."""

        return api.put(path=f'v2/tasks/{self.task_id}/{status.value}', data=data)



import itertools
from datetime import timedelta
from typing import (
    Dict,
    List,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Union,
    cast,
    overload,
)

import redis


class CacheRepositoryProtocol(Protocol):
    def set(
        self,
        key: str,
        value: str,
        ttl: int,
    ) -> None:
        ...

    def set_many(self, data: Sequence[Tuple[str, str, int]]) -> None:
        ...

    def get(self, key: str) -> Optional[str]:
        ...

    def get_many(self, keys: Sequence[str]) -> Dict[str, Optional[str]]:
        ...

    def get_all(self) -> Dict[str, str]:
        ...

    def delete(self, key: str) -> None:
        ...

    def delete_many(self, keys: Sequence[str]) -> None:
        ...

    def delete_all(self) -> None:
        ...


class RedisRepository:
    # Deletes M expired keys from the hash.
    #
    # Complexity: O(log(N) + M) + O(M) + O(M * log(N)) where N is the size of hash
    # and M is the number of elements deleted. For constant M (e.g., 3) the complexity
    # is O(log(N)).
    #
    # KEYS:
    #      hash_name.
    #      zset_name.
    #
    # ARGV:
    #      delete_count.
    #
    # Returns: nil.
    LUA_VACUUM_SCRIPT = """
    local hash_name = KEYS[1]
    local zset_name = KEYS[2]
    local delete_count = tonumber(ARGV[1])
    local time = redis.call('TIME')
    local pnow = tonumber(time[1]) * 1000 + math.floor(tonumber(time[2]) / 1000)

    local keys_to_delete = redis.call(
            'ZRANGEBYSCORE', zset_name, '-inf', pnow, 'LIMIT', 0, delete_count
    )

    if not next(keys_to_delete) then
        return
    end

    redis.call('HDEL', hash_name, unpack(keys_to_delete))
    redis.call('ZREM', zset_name, unpack(keys_to_delete))
    """

    # Gets either all or only requested keys and values from hash.
    #
    # Gets all keys if no keys specified in ARGV.
    #
    # Complexity: O(N) where N is the number of requested keys.
    #
    # KEYS:
    #     hash_name.
    #     zset_name.
    #
    # Optional ARGV:
    #     key1.
    #     key2.
    #     ...
    #
    # Returns: list of names and values for non-expired keys.
    LUA_GET_SCRIPT = """
    local hash_name = KEYS[1]
    local zset_name = KEYS[2]
    local time = redis.call('TIME')
    local pnow = tonumber(time[1]) * 1000 + math.floor(tonumber(time[2]) / 1000)

    local keys = ARGV
    if next(ARGV) == nil then
        keys = redis.call('HKEYS', hash_name)
    end

    if not next(keys) then
        return {}
    end

    local hash = redis.call('HMGET', hash_name, unpack(keys))

    local result = {}

    for i, key in ipairs(keys) do
        local pexpireat = redis.call('ZSCORE', zset_name, key)

        if not pexpireat or pnow < tonumber(pexpireat) then
            table.insert(result, key)
            table.insert(result, hash[i])
        end
    end

    return result
    """

    # Inserts list of keys-value-expiration tuples into the hash.
    #
    # Complexity: O(N * log(M)), where N is the number of inserted elements and
    # M is the hash size.
    #
    # 1. If hash does not exist, it will automatically create one.
    # 2. If the field already exists, its value and ttl will be overwritten.
    # 3. Hash and zset ttl is always set to the biggest field's ttl.
    # 4. When the field expires it may be deleted by:
    #    - Manually invoking `vacuum` script.
    #    - Redis automatically deleting expired hash (see note #3).
    #
    # KEYS:
    #     hash_name.
    #     zset_name.
    #
    # ARGV:
    #     key.
    #     value.
    #     ttl.
    #     ...
    #
    # Returns: nil.
    LUA_SET_SCRIPT = """
    local hash_name = KEYS[1]
    local zset_name = KEYS[2]
    local time = redis.call('TIME')

    for i, _ in ipairs(ARGV) do
        if i % 3 == 1 then
            local key = ARGV[i]
            local value = ARGV[i + 1]
            local ttl = ARGV[i + 2]
            local pexpireat = (
                (tonumber(time[1]) + ttl) * 1000 + math.floor(tonumber(time[2]) / 1000)
            )

            redis.call('HSET', hash_name, key, value)
            redis.call('ZADD', zset_name, pexpireat, key)

        end
    end

    local max_pexpireat = tonumber(redis.call(
            'ZREVRANGEBYSCORE', zset_name, '+inf', '-inf', 'WITHSCORES', 'LIMIT', 0, 1
    )[2])

    redis.call('PEXPIREAT', hash_name, max_pexpireat)
    redis.call('PEXPIREAT', zset_name, max_pexpireat)
    """

    # Deletes all data from hash and zset.
    #
    # Complexity: O(N) where N is the hash size.
    #
    # KEYS:
    #     hash_name.
    #     zset_name.
    #
    # Returns: nil.
    LUA_DELETE_ALL_SCRIPT = """
    local hash_name = KEYS[1]
    local zset_name = KEYS[2]

    redis.call('DEL', hash_name, zset_name)
    """

    def __init__(self, hash_name: str, client: redis.Redis, use_lua_52: bool = False):
        self.hash_name = hash_name
        self.zset_name = f'{hash_name}.EXPIREAT'
        self.client = client
        self.lua_set_many = self.client.register_script(self.LUA_SET_SCRIPT)

        lua_get_script = self.LUA_GET_SCRIPT
        if use_lua_52:
            # Hack for tests to work with fakeredis, as it uses Lua version > 5.1.
            # In Lua 5.1, unpack was a global, but in 5.2 it's been moved to
            # table.unpack.
            lua_get_script = self.LUA_GET_SCRIPT.replace('unpack', 'table.unpack')
        self.lua_get = self.client.register_script(lua_get_script)

        self.lua_vacuum = self.client.register_script(self.LUA_VACUUM_SCRIPT)
        self.lua_delete_all = self.client.register_script(self.LUA_DELETE_ALL_SCRIPT)

    def set(self, key: str, value: str, ttl: int) -> None:
        self.set_many(data=[(key, value, ttl)])

    def set_many(self, data: Sequence[Tuple[str, str, int]]) -> None:
        self.lua_set_many(
            keys=[self.hash_name, self.zset_name],
            args=list(itertools.chain.from_iterable(data)),
        )

    def get(self, key: str) -> Optional[str]:
        return self.get_many(keys=[key]).get(key)

    def get_many(self, keys: Sequence[str]) -> Dict[str, Optional[str]]:
        data = self.lua_get(keys=[self.hash_name, self.zset_name], args=list(keys))

        data = dict(zip(data[::2], data[1::2]))

        for missing_key in set(keys) - set(data):
            data[missing_key] = None

        return data

    def get_all(self) -> Dict[str, str]:
        return cast(Dict[str, str], self.get_many(keys=[]))

    def delete(self, key: str) -> None:
        self.delete_many(keys=[key])

    def delete_many(self, keys: Sequence[str]) -> None:
        self.set_many(data=[(key, '', -1) for key in keys])

    def delete_all(self) -> None:
        self.lua_delete_all(keys=[self.hash_name, self.zset_name])

    def vacuum(self, delete_count: int) -> None:
        self.lua_vacuum(keys=[self.hash_name, self.zset_name], args=[delete_count])


class DeprecatedRedisAdapter:
    """Expands basic redis functionality

    Serves the purpose of adding custom logic on top of basic redis functions
    (e.g. adding expiry to hset, which is not available out of the box).
    """

    DEFAULT_EXPIRY: timedelta = timedelta(days=60)
    REDIS_STORED_VALUE_TYPE = Union[bytes, str, int, float]

    def __init__(
        self,
        hash_name: str,
        client: redis.Redis,
    ):
        self.default_name = hash_name
        self.client = client

        self.client.ping()

    @overload
    def hset(
        self,
        name: Optional[str] = ...,
        key: str = ...,
        value: REDIS_STORED_VALUE_TYPE = ...,
        mapping: None = ...,
        expiry: Union[int, timedelta, None] = ...,
    ) -> int:
        ...

    @overload
    def hset(
        self,
        name: Optional[str] = ...,
        key: None = ...,
        value: None = ...,
        mapping: Dict[str, REDIS_STORED_VALUE_TYPE] = ...,
        expiry: Union[int, timedelta, None] = ...,
    ) -> int:
        ...

    def hset(
        self,
        name=None,
        key=None,
        value=None,
        mapping=None,
        expiry=DEFAULT_EXPIRY,
    ) -> int:
        """Stores the data in cache

        params:
         key: key, which will contain the data
         value: data to be saved
         mapping: dict of key:data pairs to be saved
         expiry: time in seconds for when data will be deleted from cache
            expiration is reset with every hset call
            expiration can be disabled by setting expiry to None
        returns: number of inserted elements
        """

        name = name or self.default_name

        n_set = self.client.hset(name=name, key=key, value=value, mapping=mapping)

        if expiry is None and self.pttl(name=name) > 0:
            self.client.persist(name=name)

        if expiry is not None:
            self.client.expire(name=name, time=expiry)

        return n_set

    def hget(
        self, key: str, name: Optional[str] = None
    ) -> Union[REDIS_STORED_VALUE_TYPE, None]:
        """Loads data from cache

        params:
         key: key to load data from
        returns: stored data
        """

        name = name or self.default_name
        return self.client.hget(name=name, key=key)

    def hgetall(
        self, name: Optional[str] = None
    ) -> Dict[str, Union[REDIS_STORED_VALUE_TYPE]]:
        """Loads all data from cache"""

        name = name or self.default_name
        return self.client.hgetall(name=name)

    def hdel(self, keys: List[str], name: Optional[str] = None) -> int:
        """Deletes some data from cache

        params:
         keys: list of keys to delete
        returns: number of deleted elements
        """

        name = name or self.default_name
        return self.client.hdel(name, *keys)

    def delete(self, *names: List[str]) -> int:
        """Deletes all data from cache"""

        delete_names = names or [self.default_name]
        return self.client.delete(*delete_names)

    def ttl(self, name: Optional[str] = None) -> int:
        """Returns the number of seconds until expiration"""

        name = name or self.default_name
        return self.client.ttl(name=name)

    def pttl(self, name: Optional[str] = None) -> int:
        """Returns the number of milliseconds until expiration"""

        name = name or self.default_name
        return self.client.pttl(name=name)

    def exists(self, *names: List[str]) -> int:
        """Returns whether there is data in cache"""

        exists_names = names or [self.default_name]
        return self.client.exists(*exists_names)



from enum import Enum


class RerunMode(str, Enum):
    REALTIME = "realtime"
    HISTORICAL = "historical"


class EventType(str, Enum):
    PARTIAL_WELL_RERUN_MERGE = "partial-well-rerun-merge"


class SourceType(str, Enum):
    DRILLING = "drilling"
    COMPLETIONS = "completions"



from typing import Optional

import pydantic

from corva.models import base, validators


class RerunDepthRange(base.CorvaBaseEvent):
    """Rerun depth range metadata.

    Attributes:
        start: start depth.
        end: end depth.
    """

    start: float
    end: float


class RerunTimeRange(base.CorvaBaseEvent):
    """Rerun time range metadata.

    Attributes:
        start: start time.
        end: end time.
    """

    start: int
    end: int

    # validators
    _set_start = pydantic.validator('start', allow_reuse=True)(validators.from_ms_to_s)
    _set_end = pydantic.validator('end', allow_reuse=True)(validators.from_ms_to_s)


class RerunTime(base.CorvaBaseEvent):
    """Rerun metadata for time event.

    Attributes:
        id: rerun id.
        range: rerun time range.
        invoke: invoke counter.
        total: total invoke count for the rerun.
    """

    id: Optional[int]  # TODO: remove optional in v2, it was added for backward comp
    range: RerunTimeRange
    invoke: int
    total: int


class RerunDepth(base.CorvaBaseEvent):
    """Rerun metadata for depth event.

    Attributes:
        id: rerun id.
        range: rerun depth range.
        invoke: invoke counter.
        total: total invoke count for the rerun.
    """

    id: Optional[int]  # TODO: remove optional in v2, it was added for backward comp
    range: RerunDepthRange
    invoke: int
    total: int



import datetime

import pydantic


class Settings(pydantic.BaseSettings):
    # api
    API_ROOT_URL: pydantic.AnyHttpUrl
    DATA_API_ROOT_URL: pydantic.AnyHttpUrl

    # cache
    CACHE_URL: str

    # logger
    LOG_LEVEL: str = 'INFO'
    LOG_THRESHOLD_MESSAGE_SIZE: int = 1000
    LOG_THRESHOLD_MESSAGE_COUNT: int = 15

    # company and app
    APP_KEY: str  # <provider-name-with-dashes>.<app-name-with-dashes>
    PROVIDER: str

    # secrets
    SECRETS_CACHE_TTL: int = int(datetime.timedelta(minutes=5).total_seconds())


SETTINGS = Settings()



import pathlib
from importlib import machinery

import setuptools

ROOT = pathlib.Path(__file__).parent
LONG_DESCRIPTION = (
    f'{(ROOT / "README.md").read_text()}'
    f'\n\n'
    f'{(ROOT / "CHANGELOG.md").read_text()}'
)

VERSION = str(
    machinery.SourceFileLoader('version', 'src/version.py').load_module().VERSION
)

CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.8',
    'Topic :: Software Development :: Libraries',
]

setuptools.setup(
    name='corva-sdk',
    author='Jordan Ambra',
    author_email="jordan.ambra@corva.ai",
    url='https://github.com/corva-ai/python-sdk',
    version=VERSION,
    classifiers=CLASSIFIERS,
    description='SDK for building Corva DevCenter Python apps.',
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    keywords='corva, sdk',
    py_modules=[file.stem for file in pathlib.Path('src').glob('*.py')],
    packages=setuptools.find_packages("src"),
    package_dir={"": "src"},
    install_requires=[
        "fakeredis[lua] >=1.4.5, <2.0.0",
        "pydantic >=1.8.2, <2.0.0",
        "redis >=3.5.3, <4.0.0",
        "requests >=2.25.0, <3.0.0",
        "urllib3 <2",  # lambda doesnt support version 2 yet
        "tenacity >=8.2.3, <9.0.0",
    ],
    python_requires='>=3.8, <4.0',
    license='The Unlicense',
    entry_points={"pytest11": ["corva = plugin"]},
)



from __future__ import annotations

import itertools
from typing import List, Optional, Union

import pydantic

from corva.api import Api
from corva.models import validators
from corva.models.base import CorvaBaseEvent, RawBaseEvent
from corva.models.rerun import RerunDepth, RerunTime
from corva.models.scheduled.scheduler_type import SchedulerType


class RawScheduledEvent(CorvaBaseEvent, RawBaseEvent):
    """Base class for raw scheduled event data.

    Attributes:
        asset_id: asset id.
        company_id: company id.
        schedule_id: schedule id.
        app_connection_id: app connection id.
        app_stream_id: app stream id.
        scheduler_type: type of scheduled event.
    """

    asset_id: int
    company_id: int = pydantic.Field(..., alias='company')
    schedule_id: int = pydantic.Field(..., alias='schedule')
    app_connection_id: int = pydantic.Field(..., alias='app_connection')
    app_stream_id: int = pydantic.Field(..., alias='app_stream')
    scheduler_type: SchedulerType
    has_secrets: bool = False

    @staticmethod
    def from_raw_event(event: Union[dict, List[List[dict]]]) -> List[RawScheduledEvent]:
        if isinstance(event, dict):
            event = [[event]]

        # flatten the event into 1d array
        flattened_event: List[dict] = list(itertools.chain(*event))

        parsed_raw_events = pydantic.parse_obj_as(
            List[RawScheduledEvent], flattened_event
        )

        events = [
            parsed_raw_event.scheduler_type.raw_event.parse_obj(sub_event)
            for parsed_raw_event, sub_event in zip(parsed_raw_events, flattened_event)
        ]

        return events

    def set_schedule_as_completed(self, api: Api) -> None:
        """Sets schedule as completed."""
        api.post(path=f'scheduler/{self.schedule_id}/completed')


class RawScheduledDataTimeEvent(RawScheduledEvent):
    """Raw data time scheduled event data.

    Attributes:
        schedule_start: Unix timestamp, when the schedule was triggered.
        start_time: left bound of the time range, covered by this event.
            Use inclusively.
        interval: time between two schedule triggers.
        rerun: rerun metadata.
    """

    schedule_start: int
    start_time: int = None  # type: ignore
    interval: float
    rerun: Optional[RerunTime] = None

    # validators
    _set_schedule_start = pydantic.validator('schedule_start', allow_reuse=True)(
        validators.from_ms_to_s
    )

    @pydantic.root_validator(pre=False, skip_on_failure=True)
    def set_start_time(cls, values: dict) -> dict:
        """Calculates start_time field."""

        values["start_time"] = int(values["schedule_start"] - values["interval"] + 1)
        return values


class RawScheduledDepthEvent(RawScheduledEvent):
    """Raw depth scheduled event data.

    Attributes:
        top_depth: start depth in ft., covered by this event. Use exclusively.
        bottom_depth: end depth in ft., covered by this event. Use inclusively.
        log_identifier: app stream log identifier. Used to scope data by stream,
            asset might be connected to multiple depth based logs.
        interval: distance between two schedule triggers.
        rerun: rerun metadata.
    """

    top_depth: float
    bottom_depth: float
    log_identifier: str
    interval: float = pydantic.Field(..., alias='depth_milestone')
    rerun: Optional[RerunDepth] = None


class RawScheduledNaturalTimeEvent(RawScheduledEvent):
    """Raw natural scheduled event data.

    Attributes:
        schedule_start: Unix timestamp, when the schedule was triggered.
        interval: time between two schedule triggers.
        rerun: rerun metadata.
    """

    schedule_start: int
    interval: float
    rerun: Optional[RerunTime] = None

    # validators
    _set_schedule_start = pydantic.validator('schedule_start', allow_reuse=True)(
        validators.from_ms_to_s
    )



import os

import pytest


@pytest.fixture(scope='function')
def app_runner():
    """Returns a function that should be used to run apps in tests."""

    # imports are local to avoid loading packages, on the first plugin run
    from corva.testing import TestClient

    return TestClient.run


@pytest.hookimpl(tryfirst=True)
def pytest_load_initial_conftests(args, early_config, parser):
    """Sets test environment variables.

    This hook is triggered before loading the packages being tested.
    If packages contain global variables, that read from env,
    they will be able to read env values set by this hook.

    Why does this hook get triggered before loading the packages being tested?
      The current file that contains the hook is registered as a pytest plugin through
      setuptools entry points. Plugins registered like this are loaded before loading
      the packages being tested.
      See load order here:
        https://docs.pytest.org/en/stable/writing_plugins.html#plugin-discovery-order-at-tool-startup
    """

    provider = 'test-provider'
    env = {
        'API_ROOT_URL': 'https://api.localhost.ai',
        'DATA_API_ROOT_URL': 'https://data.localhost.ai',
        'CACHE_URL': 'redis://localhost:6379',
        'APP_KEY': f'{provider}.test-app-name',
        'PROVIDER': provider,
        **os.environ,  # override env values if provided by user
    }
    os.environ.update(env)



def from_ms_to_s(timestamp: int) -> int:
    """Casts Unix timestamp from milliseconds to seconds.

    Casts Unix timestamp from millisecond to seconds, if provided timestamp is in
        milliseconds.
    """

    # 1 January 10000 00:00:00 - first date to not fit into the datetime instance
    if timestamp >= 253402300800:
        timestamp //= 1000

    return timestamp



import datetime
from typing import Dict, Optional, Protocol, Tuple

from corva import api


class ApiSdkProtocol(Protocol):
    def get_secrets(self, app_key: str) -> Dict[str, str]:
        ...


class CachingApiSdk:
    SECRETS_CACHE: Dict[str, Tuple[datetime.datetime, Dict[str, str]]] = {}

    def __init__(self, api_sdk: ApiSdkProtocol, ttl: int):
        self.api_sdk = api_sdk
        self.ttl = ttl

    def get_secrets(self, app_key: str) -> Dict[str, str]:
        cache_entry = self.SECRETS_CACHE.get(app_key)

        if cache_entry is not None and cache_entry[0] > datetime.datetime.now(
            tz=datetime.timezone.utc
        ):
            secrets = cache_entry[1]
        else:
            secrets = self.api_sdk.get_secrets(app_key=app_key)
            expireat = datetime.datetime.now(
                tz=datetime.timezone.utc
            ) + datetime.timedelta(seconds=self.ttl)
            self.SECRETS_CACHE[app_key] = (expireat, secrets)

        return secrets


class CorvaApiSdk:
    def __init__(self, api_adapter: api.Api):
        self.api_adapter = api_adapter

    def get_secrets(self, app_key: str) -> Dict[str, str]:
        response = self.api_adapter.get(
            path=f'/v2/apps/secrets/values?app_key={app_key}'
        )
        secrets = response.json()
        return secrets


class FakeApiSdk:
    def __init__(self, secrets: Optional[Dict[str, Dict[str, str]]] = None):
        self.secrets = secrets or {}

    def get_secrets(self, app_key: str) -> Dict[str, str]:
        return self.secrets[app_key]
